@@ -1,4 +1,4 @@
-from argparse import REMAINDER, ArgumentParser
+from argparse import ArgumentParser
 import os
 import sys
 import json
@@ -46,10 +46,6 @@ class RustBackendBase(BackendBase):
                 if ret["env_vars"]["__GRASS_CONST_BAG"] != "":
                     ret["env_vars"]["__GRASS_CONST_BAG"] += ";"
                 ret["env_vars"]["__GRASS_CONST_BAG"] += value
-        ret["use_cache"] = self._enable_cache
-        ret["update_cache"] = self._update_cache
-        if self._cache_root != None:
-            ret["cache_root"] = self._cache_root
         return ret
     def _build_crate_source(self, source):
         if source == None:
@@ -98,17 +94,8 @@ class RustBackendBase(BackendBase):
         self._runtime_source = self._build_crate_source(source)
     def set_macro_source(self, source):
         self._macro_source = self._build_crate_source(source)
-    def enable_cache(self, value = True):
-        self._enable_cache = value
-    def update_cache(self, value = True):
-        self._update_cache = value
-    def cache_root(self, value):
-        self._cache_root = value
     def __init__(self):
         super().__init__()
-        self._enable_cache = True
-        self._update_cache = True
-        self._cache_root = None
         self._const_bag = None
         self._ir_list = []
         self._argv = sys.argv[1:]
