@@ -1,18 +1,26 @@
 from abc import abstractclassmethod
 from asyncio.tasks import _unregister_task
+from functools import wraps
 from json import dumps
 from typing import Any
 
+from numpy import inner
+
 class IRBase(object):
     def __init__(self, opcode : str):
         self._opcode = opcode
-    def to_dict(self) -> dict[str]:
+    def to_dict(self, const_bag: list = None) -> dict[str]:
         ret = dict()
         ret["opcode"] = self._opcode
         return ret
     def to_json(self, indent = None, **kwargs) -> str:
         kwargs["indent"] = indent
         return dumps(self.to_dict(), **kwargs)
+    def lift_const_and_jsonify(self, indent = None, **kwargs):
+       bag = list()
+       kwargs["indent"] = indent
+       jsonified = dumps(self.to_dict(bag), **kwargs) 
+       return (jsonified, bag)
     def defs(self) -> list[str] :
         ret = []
         for key in dir(self):
@@ -44,6 +52,28 @@ class IRBase(object):
                 ret.append(item)
         return ret
 
+def make_const_bag_ref(value, bag):
+    key_id = len(bag)
+    bag.append(value)
+    return {
+        "const_bag_key": key_id
+    }
+def lift_constant_to_env(dict_value: dict[str, Any], const_bag: list):
+    if const_bag == None:
+        return
+    for key in dict_value.keys():
+        if key == "opcode":
+            continue
+        key_type = type(dict_value[key])
+        if key_type == int or key_type == float or key_type == str:
+            dict_value[key] = make_const_bag_ref(dict_value[key], const_bag)
+def try_lift_const(inner):
+    @wraps(inner)
+    def _to_dict_and_lift(self, bag : list = None):
+        ret = inner(self, bag)
+        lift_constant_to_env(ret, bag)
+        return ret
+    return _to_dict_and_lift
 
 # Actual IR representations
 
@@ -54,9 +84,9 @@ class SortedRandomInterval(IRBase):
         self._count = count
         self._min_len = min_len
         self._max_len = max_len
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["count"] = self._count
+    @try_lift_const
+    def to_dict(self, bag) -> dict[str]:
+        ret = super().to_dict(bag)
         ret["min_length"] = self._min_len
         ret["max_length"] = self._max_len
         return ret
@@ -67,11 +97,11 @@ class InlineRust(IRBase):
         super().__init__("InlineRust")
         self._env = env
         self._src = src
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
+    def to_dict(self, bag) -> dict[str]:
+        ret = super().to_dict(bag)
         ret["env"] = {}
         for key, val in self._env.items():
-            ret["env"][key] = val.to_dict()
+            ret["env"][key] = val.to_dict(bag)
         ret["src"] = self._src
         return ret
     def uses(self):
@@ -90,8 +120,9 @@ class LoadGenomeFile(IRBase):
     def __init__(self, path: str):
         super().__init__("LoadGenomeFile")
         self._path = path
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
+    @try_lift_const
+    def to_dict(self, bag) -> dict[str]:
+        ret = super().to_dict(bag)
         ret["File"] = self._path
         return ret
 
@@ -106,10 +137,10 @@ class Let(LabelAssignmentBase):
         self._value = value
     def defs(self) -> list[str]:
         return [self._id] + super().defs()
-    def to_dict(self) -> dict:
-        ret = super().to_dict()
+    def to_dict(self, bag = None) -> dict:
+        ret = super().to_dict(bag)
         ret["id"] = self._id
-        ret["value"] = self._value.to_dict()
+        ret["value"] = self._value.to_dict(bag)
         return ret
 
 class Ref(LabelAssignmentBase):
@@ -118,8 +149,8 @@ class Ref(LabelAssignmentBase):
         self._id = id
     def uses(self) -> list[str]:
         return [self._id] + super().uses()
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
+    def to_dict(self, bag) -> dict[str]:
+        ret = super().to_dict(bag)
         ret["id"] = self._id
         return ret
 
@@ -137,9 +168,14 @@ class OpenFile(BatchOperationBase):
         self._compression = compression
         self._num_of_fields = num_of_fields
         self._sorted = sorted
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["target"] = self._target
+    def to_dict(self, bag) -> dict[str]:
+        ret = super().to_dict(bag)
+        if "Path" in self._target and bag != None:
+            ret["target"] = {
+                "Path": make_const_bag_ref(self._target["Path"], bag)
+            }
+        else:
+            ret["target"] = self._target
         ret["format"] = self._format
         ret["num_of_fields"] = self._num_of_fields
         ret["compression"] = self._compression
@@ -153,9 +189,9 @@ class CastToBed(BatchOperationBase):
         self._inner = inner
         self._nof = num_of_fields
         self._sorted = sorted
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["inner"] = self._inner.to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
         ret["num_of_fields"] = self._nof
         ret["sorted"] = self._sorted
         return ret
@@ -166,10 +202,10 @@ class GroupBy(BatchOperationBase):
         super().__init__("GroupBy")
         self._inner = inner
         self._key_func = key_func
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["inner"] = self._inner.to_dict()
-        ret["keys"] = [key.to_dict() for key in self._key_func]
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
+        ret["keys"] = [key.to_dict(bag) for key in self._key_func]
         return ret
 
 class Format(BatchOperationBase):
@@ -178,23 +214,43 @@ class Format(BatchOperationBase):
         self._inner = inner
         self._fmt_str = fmt_str
         self._values = values
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["inner"] = self._inner.to_dict()
+    
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
         ret["fmt_str"] = self._fmt_str
         ret["values"] = dict()
         for key in self._values:
-            ret["values"][key] = self._values[key].to_dict()
+            ret["values"][key] = self._values[key].to_dict(bag)
         return ret
 
 class AssumeSortedIR(BatchOperationBase):
     def __init__(self, inner: IRBase):
         super().__init__("AssumeSorted")
         self._inner = inner
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["inner"] = self._inner.to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
         return ret
+
+class InternalSort(BatchOperationBase):
+    def __init__(self, base: IRBase):
+        super().__init__("InternalSort")
+        self._inner = base
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
+        return ret
+
+class Nop(BatchOperationBase):
+    def __init__(self, inner: IRBase):
+        super().__init__("Nop")
+        self._inner = inner
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
+        return ret
+
 class Alter(BatchOperationBase):
     def __init__(self, base : IRBase, target_field : str, value_expr : IRBase, sorted: bool):
         super().__init__("Alter")
@@ -202,11 +258,11 @@ class Alter(BatchOperationBase):
         self._target_field = target_field
         self._value_expr = value_expr
         self._sorted = sorted
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["inner"] = self._inner.to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
         ret["field"] = self._target_field
-        ret["value"] = self._value_expr.to_dict()
+        ret["value"] = self._value_expr.to_dict(bag)
         ret["sorted"] = self._sorted
         return ret
 
@@ -215,21 +271,50 @@ class Filter(BatchOperationBase):
         super().__init__("Filter")
         self._inner = base
         self._cond = cond
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["inner"] = self._inner.to_dict()
-        ret["cond"] = self._cond.to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
+        ret["cond"] = self._cond.to_dict(bag)
         return ret
 
-class Merge(BatchOperationBase):
-    def __init__(self, inner : IRBase, sorted : bool):
-        super().__init__("Merge")
+class Invert(BatchOperationBase):
+    def __init__(self, inner: IRBase):
+        super().__init__("Invert")
         self._inner = inner
-        self._sorted = sorted
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["inner"] = self._inner.to_dict()
-        ret["sorted"] = self._sorted
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
+        return ret
+
+class AssignTag(BatchOperationBase):
+    def __init__(self, inner: IRBase, tag):
+        super().__init__("AssignTag")
+        self._inner = inner
+        self._tag = tag
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
+        ret["tag"] = self._tag
+        return ret
+
+class TwoWayMerge(BatchOperationBase):
+    def __init__(self, a: IRBase, b: IRBase):
+        super().__init__("TwoWayMerge")
+        self._a = a
+        self._b = b
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["expr_1"] = self._a.to_dict(bag)
+        ret["expr_2"] = self._b.to_dict(bag)
+        return ret
+
+class MergeOverlap(BatchOperationBase):
+    def __init__(self, inner : IRBase):
+        super().__init__("MergeOverlap")
+        self._inner = inner
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["inner"] = self._inner.to_dict(bag)
         return ret
 
 class Intersection(BatchOperationBase):
@@ -241,11 +326,11 @@ class Intersection(BatchOperationBase):
         self._lhs = lhs
         self._rhs = rhs
         self._sorted = sorted
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
         ret["flavor"] = self._flavor
-        ret["lhs"] = self._lhs.to_dict()
-        ret["rhs"] = self._rhs.to_dict()
+        ret["lhs"] = self._lhs.to_dict(bag)
+        ret["rhs"] = self._rhs.to_dict(bag)
         ret["sorted"] = self._sorted
         return ret
 
@@ -256,19 +341,36 @@ class WriteFile(BatchOperationBase):
         super().__init__("WriteFile")
         self._what = what
         self._target = target
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["what"] = self._what.to_dict()
-        ret["target"] = self._target
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["what"] = self._what.to_dict(bag)
+        if type(self._target) == str and bag != None:
+            ret["target"] = make_const_bag_ref(self._target, bag)
+        else:
+            ret["target"] = self._target
         return ret
 
 class Count(BatchOperationBase):
     def __init__(self, what : IRBase):
         super().__init__("Count")
         self._what = what
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["what"] = self._what.to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["what"] = self._what.to_dict(bag)
+        return ret
+
+class Limit(BatchOperationBase):
+    def __init__(self, what: IRBase, count: int):
+        super().__init__("Limit")
+        self._what = what
+        self._count = count
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["what"] = self._what.to_dict(bag)
+        if bag != None:
+            ret["count"] = make_const_bag_ref(self._count, bag) 
+        else:
+            ret["count"] = self._count
         return ret
 
 ## The Field Expression
@@ -278,19 +380,19 @@ class FieldExpressionBase(IRBase):
 class RuntimeValueBase(FieldExpressionBase):
     def __init__(self, opcode : str):
         super().__init__(opcode)
-    def to_dict(self) -> dict[str]:
-        return super().to_dict()
+    def to_dict(self, bag) -> dict[str]:
+        return super().to_dict(bag)
 
 class UnaryBase(FieldExpressionBase):
     def __init__(self, opcode : str, operand_key : str, operand : IRBase):
         super().__init__(opcode)
         self._dict = dict[str, IRBase]()
         self._dict[operand_key] = operand
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
         for key in self._dict:
             if isinstance(self._dict[key], IRBase):
-                ret[key] = self._dict[key].to_dict()
+                ret[key] = self._dict[key].to_dict(bag)
             else:
                 ret[key] = self._dict[key]
         return ret
@@ -307,11 +409,11 @@ class BinaryBase(FieldExpressionBase):
         self._dict = dict()
         self._dict[lhs_key] = lhs
         self._dict[rhs_key] = rhs
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
         for key in self._dict:
             if isinstance(self._dict[key], IRBase):
-                ret[key] = self._dict[key].to_dict()
+                ret[key] = self._dict[key].to_dict(bag)
             else:
                 ret[key] = self._dict[key]
         return ret
@@ -324,11 +426,11 @@ class Cond(FieldExpressionBase):
         self._cond = cond
         self._then = then
         self._else = elze
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
-        ret["cond"] = self._cond.to_dict()
-        ret["then"] = self._then.to_dict()
-        ret["elze"] = self._else.to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
+        ret["cond"] = self._cond.to_dict(bag)
+        ret["then"] = self._then.to_dict(bag)
+        ret["elze"] = self._else.to_dict(bag)
         return ret
 
 class FieldRef(UnaryBase):
@@ -344,8 +446,8 @@ class ComponentFieldRef(FieldExpressionBase):
         super().__init__("ComponentFieldRef")
         self._target = target
         self._field_name = field_name
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
+    def to_dict(self, bag = None) -> dict[str]:
+        ret = super().to_dict(bag)
         ret["target"] = self._target
         ret["field_name"] = self._field_name
         return ret
@@ -353,6 +455,12 @@ class ComponentFieldRef(FieldExpressionBase):
 class ConstValue(UnaryBase):
     def __init__(self, value : Any):
         super().__init__("ConstValue", "value", value)
+    def to_dict(self, bag) -> dict[str]:
+        ret = super().to_dict(bag)
+        type_of_value = type(self._dict["value"])
+        if bag != None and (type_of_value in [int, float, str]):
+            ret["value"] = make_const_bag_ref(self._dict["value"], bag)
+        return ret
 
 class FullRecordRef(RuntimeValueBase):
     def __init__(self):
@@ -362,8 +470,8 @@ class RecordRef(IRBase):
     def __init__(self, id : int):
         super().__init__("RecordRef")
         self._id = id
-    def to_dict(self) -> dict[str]:
-        ret = super().to_dict()
+    def to_dict(self, bag) -> dict[str]:
+        ret = super().to_dict(bag)
         ret["id"] = self._id
         return ret
 
